'''
This is the Offer class, which can be built from an XML file in order to
provide verification.
'''

from xml.etree.ElementTree import ElementTree
from repr import rtoi
import datetime
import string

debug_flag = 0
OFF_TYPE_SECONDARY = 0
OFF_TYPE_TERTIARY_UP = 1
OFF_TYPE_TERTIARY_DOWN = 2

OFF_TYPE_SALE = 0
OFF_TYPE_BUY = 1
OFFER_GRANULARITY = 60

def debug(str):
    if debug_flag:
        print str

class Offer:
    # Configurable values
    num_intervals = 24
    max_off_per_interval = 10;

    def add_error(self, message):
##
# Adds an error to the error stack.
# @ param message The error message to be added in the stack.
# @return Nothing
        debug(message)
        self.error_stack.append(message)

    def __init__(self, xml_file, msg_timestamp):
##
# Builds an Offer class from a given XML file
#
# @param xml_file The name of the XML file from which the class is to be
# initialized
# @param timestamp The timestamp of the message
# @return an Offer object
# @defreturn Offer
# @exception ParseError raised by the parser if it cannot load the file
# @exception IndexError raised by tree.find if it cannot find an element
# @exception ? raised by datetime.date.strptime()

        self.error_stack = []
        self.energy_offer = []

        self.tree = ElementTree()
        self.tree.parse(xml_file)

        self.sender_identification = self.tree.find('SenderIdentification').get('v')
        self.receiver_identification = self.tree.find('ReceiverIdentification').get('v')

        self.message_version = int(self.tree.find("MessageVersion").get('v'))
        self.message_date_time = datetime.datetime.now()

        offer_date_str = self.tree.find("OfferDate").get("v")
        if (offer_date_str):
            self.offer_date = datetime.datetime.strptime(offer_date_str, "%Y-%d-%m")

        self.resoluion = 1

        self.energy_type = self.tree.find("EnergyType").get('v')
        if (self.energy_type != None):
            self.energy_type = self.energy_type.lower()

        offer_type = self.tree.find("OfferType").get('v')
        if (offer_type != None):
            if (offer_type.lower() == "sell"):
                self.offer_type = OFF_TYPE_SALE
            elif (offer_type.lower() == "buy"):
                self.offer_type = OFF_TYPE_BUY
            else:
                self.add_error("Unknown offer type")

        self.energy_offer_element = self.tree.find("EnergyOffer")

        if (self.energy_offer_element == None):
            self.add_error("No energy offer")
            return

        intervals = self.energy_offer_element.findall("Interval")
        debug("===\n")

        if (intervals == []):
            self.add_error("No intervals found")
            return
        debug("Got intervals\n")
        debug(intervals)

        self.earliest_interval_offered = None

        for interval in intervals:
            idx = int(interval.get("v"))
            offers = interval.findall("Block")
            if (offers == [] or idx == None):
                self.add_error("No offer or index for interval " + str(idx))
                return
            for offer in offers:
                price = rtoi(offer.find("Price").get("v"))
                qty = rtoi(offer.find("Qty").get("v"))
                if (price == None or qty == None):
                    self.add_error("No price or qty")
                    return
                self.energy_offer.append((idx, price, qty))
            if self.earliest_interval_offered == None:
                self.earliest_interval_offered = idx

    def get_earliest_interval_offered(self):
        return self.earliest_interval_offered

    def get_creation_date(self):
        return self.message_date_time

    def ger_message_version(self):
        return self.message_version

    def get_delivery_date(self):
        return self.offer_date

    def get_energy_type(self):
##
# Returns the energy type for which the current offer is placed.

        if (self.energy_type == "secondary"):
            return OFF_TYPE_SECONDARY
        if (self.energy_type == "tertiary_upward"):
            return OFF_TYPE_TERTIARY_UP
        if (self.energy_type == "tertiary_downward"):
            return OFF_TYPE_TERTIARY_DOWN

    def offer_type(self):
##
# Returns the offer type of the current offer
        return self.offer_type

    def bids_for_interval(self, interval):
##
# Returns the bids placed for a given interval.
# @param interval The interval for which the query is being made
# @return A list of bids for the required interval

        return [bid for bid in self.energy_offer if bid[0] == interval]

    def is_valid(self, check_date, max_amount, min_amount, max_price,
                 min_price, is_day_offer = False, grace = None, day_dt = None):
        ##
        # Checks if an offer is valid in terms of content
        # Verifications:
        # - The offer is at most for one day before check_date *or*
        # - The offer is ont he same day as check_date but no more than
        #   <grace> hours before 24:00
        # - There are no more than 10 bids for each interval
        # - Each bid is between max_amount and min_amount MW
        # - Each bid has a cost between min_price and max_price

        if len(self.error_stack) != 0:
            return (False, self.error_stack)

        # Is the offer for no less than 0 days before the checkdate?

        debug("Date checking:")
        debug(check_date)
        debug(self.offer_date)
        debug((check_date - self.offer_date).days)

        if ((check_date - self.offer_date).days > 0):
            self.add_error("Bidding time has passed")

        # If the offer is a day offer, are the following conditions met:
        # - Is the earliest bid at least

        if ((check_date - self.offer_date).days == 0):
            if grace == None:
                self.add_error("Day offers are not allowed for this transaction")
            else:
                ei = self.earliest_interval_offered
                if (datetime.datetime.now().hour + 1 + grace) < ei:
                    self.add_error("Bidding time for day offers has passed")

        # Are there no more than 10 bids for each interval?
        for i in range(1, self.num_intervals + 1):
            if (len(self.bids_for_interval(i)) > self.max_off_per_interval):
                print len(self.bids_for_interval(i))
                self.add_error("Too many bids for interval " + str(i))

        # Is each bid's qty between min_amount and max_amount, and price
        # between min_price and max_price?

        for i in range(1, self.num_intervals + 1):
            bids = self.bids_for_interval(i)
            for bid in bids:
                if ( (bid[1] > max_price) or (bid[1] < min_price) ):
                    self.add_error("Interval " + str(i) + " contains a " +
                                   "bid with a price outside the allowed range")
                if ( (bid[2] > max_amount) or (bid[2] < min_amount) ):
                    self.add_error("Interval " + str(i) + " contains a bid " +
                                   "with a quantity outisde the allowed range")
        if len(self.error_stack) == 0:
            return (True, [])
        else:
            return (False, self.error_stack)
